///////////////////////////////////
//                               //
//    ***   3DEng 01  ***        //
//                               //
//   Copyright  L. Polok (Mr.)  //
//                               //
//        3D Engine 01           //
//                               //
///////////////////////////////////

#include <stdlib.h>
#include <windows.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>

#define PI 3.14159265358979
#define WINDOWTITLE "3D Engine 01 - Esc to quit"
#define APPNAME "3D Engine 01"

struct Vertex {
    float x, y, z;
};

typedef float Matrix[4][4];

struct Object {
    int n_vertex_num;
    Vertex *vertex;
    Matrix m_object;
};

struct World {
	int n_object_num;
	Object *object;
};

// Promnn
bool b_loaded;         //  pomocna promenna
World *world;          //  svet
Object *svet;          //  vlastni svet (ukazuje do world->object)
Object *plane;         //  letadlo (ukazuje do world->object)
Matrix cmatrix;        //  kamera

// Program

//---------------------------------------------------
// vytvori jednotkovou matici

void Matrix_Init(Matrix *m)
{
    int i, j;

    for(i = 0; i < 4; i ++)
        for(j = 0; j < 4; j ++)
            (*m)[i][j] = (float)(i == j);
}

//---------------------------------------------------
// tato funkce precte z disku objekt

int Count_Objects(FILE *ASCfp, char *id_type)
{
    int n_objects;
    char str[256];

    fseek(ASCfp, SEEK_SET, 0);
    // na zacatek

    n_objects = 0;
    while(fgets(str, 255, ASCfp)/*!feof(ASCfp)*/) {
        
        if(strstr(str, "Named object") != NULL) {
			fgets(str, 255, ASCfp);
			if(strstr(str, id_type) != NULL)
				n_objects ++;
		}
    }
    // pocita objekty

    return n_objects;
}

Vertex Read_Vertex(FILE *ASCfp, int b_mapped)
{
    char str[256], *p_chr;
    Vertex vertex;

    fgets(str, 255, ASCfp);
    // precte radek

    if((p_chr = strstr(str, "X:")) == NULL)
        vertex.x = 0;
    else
        vertex.x = (float)atof(&p_chr[2]);
    // X

    if((p_chr = strstr(str, "Y:")) == NULL)
        vertex.z = 0;
    else
        vertex.z = (float)atof(&p_chr[2]); //
    // Y

    if((p_chr = strstr(str, "Z:")) == NULL)
        vertex.y = 0;
    else
        vertex.y = (float)atof(&p_chr[2]); // 3d-studio ma od naseho engine odlisne osy
    // Z

    return vertex;
}

int Read_Objects(FILE *ASCfp, World *world)
{
    char str[256], buff[256], *p_chr;
    int n_cur_obj, b_mapped;
    int i;

    fseek(ASCfp, SEEK_SET, 0);
    // na zacatek

    n_cur_obj = 0;
    while(fgets(str, 255, ASCfp)/*!feof(ASCfp)*/) {
        if((p_chr = strstr(str, "Named object:")) != NULL) {
            // pokud najde objekt

            p_chr += strlen("Named object: _"); // preskoci "Named object: ""
            strcpy(buff, p_chr);
            while(buff[strlen(buff) - 1] == '\"' ||
                  buff[strlen(buff) - 1] == '\n')
                buff[strlen(buff) - 1] = 0; // odkrouhne uvozovku na konci
            // vyparsuje jeho jmeno

            Matrix_Init(&world->object[n_cur_obj].m_object);
            // nastavi matici

            fgets(str, 255, ASCfp);
			if(strstr(str, "Tri-mesh") == NULL)
				continue;
			// pokud neni tri-mesh

            if((p_chr = strstr(str, "Vertices:")) == NULL)
                return 0;
            p_chr += strlen("Vertices: ");
            world->object[n_cur_obj].n_vertex_num = atoi(p_chr);
            // precte pocet vertexu

            b_mapped = 0;
            fgets(str, 255, ASCfp);
            if(!stricmp(str, "Mapped\n")) {
                b_mapped = 1;
                fgets(str, 255, ASCfp);
            }
			if(stricmp(str, "Vertex list:\n"))
                return 0;
            // precte radky mapped a vertex-list
            // (a tim zjisti zda je na nem textura)

            if((world->object[n_cur_obj].vertex = (Vertex*)malloc
               (world->object[n_cur_obj].n_vertex_num * sizeof(Vertex))) == NULL)
                return 0;
            // naalokuje pamet pro vertexy

            for(i = 0; i < world->object[n_cur_obj].n_vertex_num; i ++)
                world->object[n_cur_obj].vertex[i] = Read_Vertex(ASCfp, b_mapped);
            // precte vertexy (b_mapped se ted nepouzije)

            n_cur_obj ++;
            // zvysi pocet objektu
        }
    }
    // pocita a cte materialy

    return 1;
}

World *Load_ASC(char *name)
{
    FILE *ASCfp;
    World *tmp;

    if((ASCfp = fopen(name, "r")) == NULL)
        return NULL;
    // open file

    if((tmp = (World*)malloc(sizeof(World))) == NULL)
        return NULL;
    // alloc world

    if((tmp->n_object_num = Count_Objects(ASCfp, "Tri-mesh")) == -1)
        return NULL;
    // pocita objekty

    if((tmp->object = (Object*)malloc
       (tmp->n_object_num * sizeof(Object))) == NULL)
        return NULL;
    // naalokuje pamet pro objekty

    if(!Read_Objects(ASCfp, tmp))
        return NULL;
    // cte objekty
    
    fclose(ASCfp);
    // zavre soubor

    return tmp;
}

//---------------------------------------------------
// nasobeni dvou matic

void Matrix_Multiply(Matrix *a, Matrix *b, Matrix *r)
{
    Matrix temp;
    int i, j;

    for(i = 0; i < 4; i ++) {
        for(j = 0; j < 4; j ++)
            temp[i][j] = (*a)[0][j] * (*b)[i][0] + (*a)[1][j] * (*b)[i][1] +
                (*a)[2][j] * (*b)[i][2] + (*a)[3][j] * (*b)[i][3];
    }
    memcpy(r, temp, sizeof(Matrix));
}

//---------------------------------------------------
// rotace matice po ose x

void Matrix_Rotate_X(Matrix *m, float angle)
{
    Matrix temp;

    Matrix_Init(&temp);

    temp[2][1] = (float)-sin(angle);
    temp[2][2] = (float)cos(angle);
    temp[1][1] = (float)cos(angle);
    temp[1][2] = (float)sin(angle);

    Matrix_Multiply(m, &temp, m);
}

//---------------------------------------------------
// rotace matice po ose y

void Matrix_Rotate_Y(Matrix *m, float angle)
{
    Matrix temp;

    Matrix_Init(&temp);

    temp[0][0] = (float)cos(angle);
    temp[0][2] = (float)-sin(angle);
    temp[2][2] = (float)cos(angle);
    temp[2][0] = (float)sin(angle);

    Matrix_Multiply(m, &temp, m);
}

//---------------------------------------------------
// rotace matice po ose z

void Matrix_Rotate_Z(Matrix *m, float angle)
{
    Matrix temp;

    Matrix_Init(&temp);

    temp[0][0] = (float)cos(angle);
    temp[0][1] = (float)sin(angle);
    temp[1][1] = (float)cos(angle);
    temp[1][0] = (float)-sin(angle);

    Matrix_Multiply(m, &temp, m);
}

//---------------------------------------------------
// posunuti matice

void Matrix_Translate(Matrix * m, float dx, float dy, float dz)
{
    Matrix temp;

    Matrix_Init(&temp);

    temp[3][0] = dx;
    temp[3][1] = dy;
    temp[3][2] = dz;

    Matrix_Multiply(m, &temp, m);
}

//---------------------------------------------------
// spocita inverzni matici
void Matrix_Inverse(Matrix *src, Matrix *inv)
{
	float f_det;
	int i, j;

	f_det = (*src)[0][0] * (*src)[1][1] * (*src)[2][2] +
		    (*src)[1][0] * (*src)[2][1] * (*src)[0][2] +
			(*src)[0][1] * (*src)[1][2] * (*src)[2][0] -
			(*src)[0][2] * (*src)[1][1] * (*src)[2][0] -
			(*src)[0][1] * (*src)[1][0] * (*src)[2][2] -
			(*src)[0][0] * (*src)[1][2] * (*src)[2][1];
	// spocita determinant podle sarrusova pravidla
	// u matice se spodkem "0 0 0 1" je determinant
	// stejny jako posledni doplnek

	(*inv)[0][0] = (*src)[1][1] * (*src)[2][2] /** (*src)[3][3] +*/ // 33 = 1
		           /*(*src)[2][1] * (*src)[3][2] * (*src)[1][3] +*/ // 13 = 0
				   /*(*src)[1][2] * (*src)[2][3] * (*src)[3][1] -*/ // 23 = 0
				   /*(*src)[3][1] * (*src)[2][2] * (*src)[1][3] -*/ // 13 = 0
				   /*(*src)[1][1] * (*src)[2][3] * (*src)[3][2] -*/ // 23 = 0
				   -(*src)[1][2] * (*src)[2][1] /** (*src)[3][3]*/; // 33 = 1
	(*inv)[1][0] =-(*src)[1][0] * (*src)[2][2] /** (*src)[3][3] +*/ // 33 = 1
		           /*(*src)[2][0] * (*src)[3][2] * (*src)[1][3] +*/ // 13 = 0
				   /*(*src)[1][2] * (*src)[2][3] * (*src)[3][0] -*/ // 23 = 0
				   /*(*src)[1][3] * (*src)[2][2] * (*src)[3][0] -*/ // 13 = 0
				   /*(*src)[2][3] * (*src)[3][2] * (*src)[1][0] -*/ // 23 = 0
				   +(*src)[1][2] * (*src)[2][0] /** (*src)[3][3]*/; // 33 = 1
	(*inv)[2][0] = (*src)[1][0] * (*src)[2][1] /** (*src)[3][3] +*/ // 33 = 1
		           /*(*src)[1][3] * (*src)[2][0] * (*src)[3][1] +*/ // 13 = 0
				   /*(*src)[2][3] * (*src)[1][1] * (*src)[3][1] -*/ // 23 = 0
				   /*(*src)[1][3] * (*src)[2][1] * (*src)[3][0] -*/ // 13 = 0
				   /*(*src)[2][3] * (*src)[3][2] * (*src)[1][0] -*/ // 23 = 0
				   -(*src)[1][1] * (*src)[2][0] /** (*src)[3][3]*/; // 33 = 1
	(*inv)[3][0] =-(*src)[1][0] * (*src)[2][1] * (*src)[3][2] -
		           (*src)[1][1] * (*src)[2][2] * (*src)[3][0] -
				   (*src)[2][0] * (*src)[3][1] * (*src)[1][2] +
				   (*src)[1][2] * (*src)[2][1] * (*src)[3][0] +
				   (*src)[1][1] * (*src)[2][0] * (*src)[3][2] +
				   (*src)[2][2] * (*src)[1][0] * (*src)[3][1];
	// takhle moc se to vecko zkrt, dl u to nebudu rozepisovat ...

	(*inv)[0][1] =-(*src)[0][1] * (*src)[2][2] + (*src)[0][2] * (*src)[2][1];
	(*inv)[1][1] = (*src)[0][0] * (*src)[2][2] - (*src)[0][2] * (*src)[2][0];
	(*inv)[2][1] =-(*src)[0][0] * (*src)[2][1] + (*src)[0][1] * (*src)[2][0];
	(*inv)[3][1] = (*src)[0][0] * (*src)[2][1] * (*src)[3][2] +
		           (*src)[0][1] * (*src)[2][2] * (*src)[3][0] +
				   (*src)[2][0] * (*src)[3][1] * (*src)[0][2] -
				   (*src)[0][2] * (*src)[2][1] * (*src)[3][0] -
				   (*src)[0][1] * (*src)[2][0] * (*src)[3][2] -
				   (*src)[2][2] * (*src)[0][0] * (*src)[3][1];
	(*inv)[0][2] = (*src)[0][1] * (*src)[1][2] - (*src)[0][2] * (*src)[1][1];
	(*inv)[1][2] =-(*src)[0][0] * (*src)[1][2] + (*src)[0][2] * (*src)[1][0];
	(*inv)[2][2] = (*src)[0][0] * (*src)[1][1] - (*src)[0][1] * (*src)[1][0];
	(*inv)[3][2] =-(*src)[0][0] * (*src)[1][1] * (*src)[3][2] -
		           (*src)[0][1] * (*src)[1][2] * (*src)[3][0] -
				   (*src)[1][0] * (*src)[3][1] * (*src)[0][2] +
				   (*src)[0][2] * (*src)[1][1] * (*src)[3][0] +
				   (*src)[0][1] * (*src)[1][0] * (*src)[3][2] +
				   (*src)[1][2] * (*src)[0][0] * (*src)[3][1];

	for(j = 0; j < 4; j ++)
		for(i = 0; i < 3; i ++)
			(*inv)[j][i] /= f_det;
	// vecko vydl determinantem (krom poslednho sloupce)

	(*inv)[0][3] = 0;
	(*inv)[1][3] = 0;
	(*inv)[2][3] = 0;
	(*inv)[3][3] = 1;
	// pi matici se spodnm dkem "0 0 0 1" vyjde vdycky posledn sloupec takhle.
}

LRESULT CALLBACK Wokno (HWND, UINT, WPARAM, LPARAM);

//---------------------------------------------------
// main
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdline, int nCmdShow)
{
	static char szAppName[] = APPNAME;
	HWND		hwnd;
	MSG			msg;
	WNDCLASSEX	wndclass;

	wndclass.cbSize			= sizeof (wndclass);
	wndclass.style			= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= Wokno;
	wndclass.cbClsExtra		= 0;
	wndclass.cbWndExtra		= 0;
	wndclass.hInstance		= hInstance;
  	wndclass.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_APPLICATION));
  	wndclass.hCursor		= LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW));
	wndclass.hbrBackground	= (HBRUSH) GetStockObject (WHITE_BRUSH);
	wndclass.lpszMenuName	= 0;
	wndclass.lpszClassName	= szAppName;
	wndclass.hIconSm		= LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
	RegisterClassEx(&wndclass);

	b_loaded = false;
	// promnn, kter ukazuje jestli se u svt nahrl
	// je to kvli tomu e kreslen je ve WM_PAINT, kter probhne
	// u po ShowWindow(), tedy jet ne se cokoli nahraje
	
	hwnd = CreateWindowEx(WS_EX_TOPMOST,
		szAppName,
		WINDOWTITLE,
		WS_DLGFRAME,
		CW_USEDEFAULT, CW_USEDEFAULT,
		320, 200,
		NULL,
		NULL,
		hInstance,
		NULL);
	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);
	// Vytvo okno

	world = Load_ASC("Svet.asc");
	if(world == NULL || world->n_object_num != 2)
		return NULL;
	// nahraje svt

	plane = &(world->object[1]);
	svet = &(world->object[0]);
	// nastav pointery na objekty

	Matrix_Init(&cmatrix);
	Matrix_Rotate_Y(&cmatrix, (float)-PI / 6);
	Matrix_Rotate_X(&cmatrix, (float)PI / 5);
	Matrix_Translate(&cmatrix, 0, 0, -500);
	// vytvo matici kamery

	b_loaded = true;
	// vse je hotovo

	while(GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}

	free((void*)svet->vertex);
	// uvolni z pameti svet

	free((void*)plane->vertex);
	// uvolni z pameti letadlo

	free((void*)world->object);
	free((void*)world);
	// uvolni strukturu sveta

	return msg.wParam;
}

//---------------------------------------------------
// pi stisku klvesy posune matici
void kdown(int kcode)
{
	if(kcode == VK_UP)
		Matrix_Rotate_X(&cmatrix, (float)PI/80);
	if(kcode == VK_DOWN)
		Matrix_Rotate_X(&cmatrix, (float)-PI/80);
	if(kcode == VK_LEFT)
		Matrix_Rotate_Y(&cmatrix, (float)-PI/80);
	if(kcode == VK_RIGHT)
		Matrix_Rotate_Y(&cmatrix, (float)PI/80);
	if(kcode == VK_INSERT)
		Matrix_Translate(&cmatrix, 0, 0, 5);
	if(kcode == VK_HOME)
		Matrix_Rotate_Z(&cmatrix, (float)-PI/80);
	if(kcode == VK_DELETE)
		Matrix_Translate(&cmatrix, 0, 0, -5);
	if(kcode == VK_END)
		Matrix_Rotate_Z(&cmatrix, (float)PI/80);
	// pohyb kamerou

	if(kcode == 104) // numpad_up
		Matrix_Rotate_X(&plane->m_object, (float)PI/80);
	if(kcode == 98)  // numpad_down
		Matrix_Rotate_X(&plane->m_object, (float)-PI/80);
	if(kcode == 100) // numpad_left
		Matrix_Rotate_Y(&plane->m_object, (float)-PI/80);
	if(kcode == 102)// numpad_right
		Matrix_Rotate_Y(&plane->m_object, (float)PI/80);
	if(kcode == 107) // numpad +
		Matrix_Translate(&plane->m_object, 0, 0, 5);
	if(kcode == 105) // numpad 9
		Matrix_Rotate_Z(&plane->m_object, (float)-PI/80);
	if(kcode == 109) // numpad -
		Matrix_Translate(&plane->m_object, 0, 0, -5);
	if(kcode == 103) // numpad 7
		Matrix_Rotate_Z(&plane->m_object, (float)PI/80);
	// pohyb letadlem
}

//---------------------------------------------------
// kresli object obj z pohledu kamery cam do hdc

void DrawObject(Object *p_object, Matrix *p_camera, HDC hdc)
{
	float x, y, z, tx, ty, tz;
	int num, i, j;
	Matrix m;
	// promenne

	Matrix_Inverse(p_camera,&m);
	Matrix_Multiply(&m, &p_object->m_object, &m);
	// vytvori novou matici

	// kresli vertexy
	for(num = 0; num < p_object->n_vertex_num; num ++) {
		x = p_object->vertex[num].x;
		y = p_object->vertex[num].y;
		z = p_object->vertex[num].z;

		tx = x * m[0][0] + y * m[1][0] + z * m[2][0] + m[3][0];
		ty = x * m[0][1] + y * m[1][1] + z * m[2][1] + m[3][1];
		tz = x * m[0][2] + y * m[1][2] + z * m[2][2] + m[3][2];
		// transformace

		if(tz > 0.1) {
			i = (int)(160 + (256 * tx) / tz);
			j = (int)(100 - (256 * ty) / tz);
			// perspektivni korekce (z_delta je tam "od oka")

			if(i >= 0 && i < 320 && j >= 0 && j < 200) {
				MoveToEx(hdc, i, j, NULL);
				LineTo(hdc, i + 1, j + 1);
			} // kresleni
		}
	}
}


LRESULT CALLBACK Wokno(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	HDC	hdc;
	PAINTSTRUCT ps;
	RECT rect;
	HBRUSH h = CreateSolidBrush(RGB(0, 0, 0));

	switch (iMsg)
	{
		case WM_PAINT :
			if(b_loaded == false)
				return 0;
			// pokud jeste neni nahrany svet

			hdc = BeginPaint (hwnd, &ps);	
			SelectObject(hdc, &h);
			// pripravi ke kresleni

			Rectangle(hdc, 0, 0, 320, 200);
			// nakresli obdelnik pres cely obraz
			
			DrawObject(svet, &cmatrix, hdc);
			DrawObject(plane, &cmatrix, hdc);
			// kresli objekty
			
			EndPaint(hwnd, &ps);
			// ukonci kresleni

			return 0;

		case WM_KEYDOWN :
			GetClientRect(hwnd, &rect);
			if(wParam == 27) {
				PostMessage(hwnd, WM_CLOSE, 0, 0);
				// ukonci program (27 = esc)
			} else {
				kdown(wParam);
				InvalidateRect(hwnd, &rect, TRUE);
			} // posune kameru a prekresli obraz
			
			return 0 ; 

		case WM_DESTROY :
			PostQuitMessage (0);
			return 0 ;
	}

	return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}